// ChatServlet - backend for ACME Chat
//
// Copyright (C) 1996 by Jef Poskanzer <jef@acme.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit the ACME Labs Java page for up-to-date versions of this and other
// fine Java utilities: http://www.acme.com/java/

package Acme.Chat;

import java.io.*;
import java.util.*;
import Acme.Serve.*;
import Acme.Serve.servlet.*;
import Acme.Serve.servlet.http.*;

/// Backend for ACME Chat.
// <P>
// <A HREF="/resources/classes/Acme/Chat/ChatServlet.java">Fetch the software.</A><BR>
// <A HREF="/resources/classes/Acme.tar.Z">Fetch the entire Acme package.</A>

public class ChatServlet extends HttpServlet
    {

    private static final String progName = "ChatServlet";

    public static void main( String[] args )
	{
	// Parse args.
	int port = 9090;
	int argc = args.length;
	int argn;
	for ( argn = 0; argn < argc && args[argn].charAt( 0 ) == '-'; ++argn )
	    {
	    if ( args[argn].equals( "-p" ) && argn + 1 < argc )
		{
		++argn;
		port = Integer.parseInt( args[argn] );
		}
	    else
		usage();
	    }
	if ( argn != argc )
	    usage();

	// Create the server.
	Serve serve = new Serve( port, System.err );

	// Add our custom Servlets.
	serve.addServlet( "/servlet/ChatServlet", new Acme.Chat.ChatServlet() );
	serve.addServlet( "/servlet/TestServlet", new Acme.Serve.TestServlet() );

	// And add the defauls Servlets, so ChatApplet can get fetched.
	serve.addDefaultServlets( false );

	// And run.
	serve.serve();

	// If we get here we're done.
	System.exit( 0 );
	}

    private static void usage()
	{
	System.err.println( "usage:  " + progName + " [-p port]" );
	System.exit( 1 );
	}


    /// Returns a string containing information about the author, version, and
    // copyright of the servlet.
    public String getServletInfo()
	{
	return getClass().getName() +
 " - Copyright (C) 1996 by Jef Poskanzer <jef@acme.com>.  All rights reserved.";
	}


    private Random random = new Random();

    private String[] lines = new String[ChatUtils.NUM_LINES];
    private int lastLine = -1;
    private Hashtable userTable = new Hashtable();

    /// Services a single request from the client.
    // @param req the servlet request
    // @param req the servlet response
    // @exception ServletException when an exception has occurred
    public void service( HttpServletRequest req, HttpServletResponse res ) throws ServletException, IOException
        {
	expire();

	String userid = req.getParameter( ChatUtils.PRM_USERID );
	String command = req.getParameter( ChatUtils.PRM_COMMAND );
	if ( userid == null || command == null )
	    {
	    res.sendError(
		HttpServletResponse.SC_BAD_REQUEST, "missing parameters" );
	    return;
	    }

	ChatServletUser user = (ChatServletUser) userTable.get( userid );
	if ( user == null )
	    {
	    String host = req.getRemoteHost();
	    if ( host == null )
		host = req.getRemoteAddr();
	    user = new ChatServletUser( userid, "anon", host, lastLine );
	    userTable.put( userid, user );
	    }

	String method = req.getMethod();
	if ( method.equalsIgnoreCase( "get" ) )
	    serviceGet( req, res, user, command );
	else if ( method.equalsIgnoreCase( "post" ) )
	    servicePost( req, res, user, command );
	else
	    res.sendError(
		HttpServletResponse.SC_BAD_REQUEST, "unknown method" );
	}


    private void serviceGet( HttpServletRequest req, HttpServletResponse res, ChatServletUser user, String request ) throws ServletException, IOException
	{
	if ( request.equals( ChatUtils.REQ_GET_LINE ) )
	    getLine( req, res, user );
	else if ( request.equals( ChatUtils.REQ_WHOS_ON ) )
	    whosOn( req, res, user );
	else
	    res.sendError(
		HttpServletResponse.SC_BAD_REQUEST, "unknown request" );
	}
    
    private synchronized void getLine( HttpServletRequest req, HttpServletResponse res, ChatServletUser user ) throws IOException
	{
	log( "getting line for " + user.nameAtHost );
	long start = System.currentTimeMillis();
	while ( true )
	    {
	    // Is a new line available?
	    if ( lastLine != user.lastLineSeen )
		{
		// Got one.
		user.lastLineSeen = ( user.lastLineSeen + 1 ) % lines.length;
		sendResponse( res, ChatUtils.RES_OK, lines[user.lastLineSeen] );
		return;
		}

	    // Has the timeout elapsed?
	    long now = System.currentTimeMillis();
	    long elapsed = now - start;
	    if ( elapsed >= ChatUtils.INT_TIMEOUT )
		break;

	    // Sleep till the timeout or a notify.
	    try
		{
		wait( ChatUtils.INT_TIMEOUT - elapsed );
		}
	    catch ( InterruptedException ignore ) {}
	    }

	// Timeout.  Return nothing.
	sendResponse( res, ChatUtils.RES_NOTHING, "" );
	}

    private void whosOn( HttpServletRequest req, HttpServletResponse res, ChatServletUser user ) throws IOException
	{
	user.active();
	log( "who's on by " + user.nameAtHost );
	long now = System.currentTimeMillis();
	StringBuffer who = new StringBuffer(
	    "  name                            idle\n" );
	Enumeration en = userTable.keys();
	while ( en.hasMoreElements() )
	    {
	    String userid = (String) en.nextElement();
	    ChatServletUser u = (ChatServletUser) userTable.get( userid );
	    if ( u.connected )
		{
		who.append( "  " );
		who.append( Acme.Fmt.fmt( u.nameAtHost, 30, Acme.Fmt.LJ ) );
		who.append( "  " );
		who.append( Acme.Utils.intervalStr( now - u.lastTimeActive ) );
		who.append( "\n" );
		}
	    }
	sendResponse( res, ChatUtils.RES_OK, who.toString() );
	}

    private void sendResponse( HttpServletResponse res, String command, String data ) throws IOException
	{
	res.setContentType( "text/plain" );
	res.setContentLength( command.length() + 1 + data.length() );
	ServletOutputStream out = res.getOutputStream();
        out.println( command );
        out.print( data );
	out.close();
        }


    private void servicePost( HttpServletRequest req, HttpServletResponse res, ChatServletUser user, String command ) throws ServletException, IOException
	{
	int len = req.getContentLength();
	if ( len == -1 )
	    res.sendError(
		HttpServletResponse.SC_BAD_REQUEST, "malformed POST" );
	else
	    {
	    ServletInputStream in = req.getInputStream();
	    byte[] bytes = new byte[len];
	    if ( Acme.Utils.readFully( in, bytes, 0, len ) == -1 )
		throw new IOException( "incomplete POST data" );
	    in.close();
	    String data = new String( bytes, 0, 0, len );
	    if ( command.equals( ChatUtils.CMD_CONNECT ) )
		connect( user, data );
	    else if ( command.equals( ChatUtils.CMD_NEW_NAME ) )
		newName( user, data );
	    else if ( command.equals( ChatUtils.CMD_NEW_LINE ) )
		newLine( user, data );
	    else if ( command.equals( ChatUtils.CMD_DISCONNECT ) )
		disconnect( user, data );
	    else
		res.sendError(
		    HttpServletResponse.SC_BAD_REQUEST,
		    "unknown command" + command );
	    }
	}

    private void connect( ChatServletUser user, String data )
	{
	user.active();
	log( "connected: " + user.nameAtHost );
	addLine( user.nameAtHost + " has connected." );
	}

    private void newName( ChatServletUser user, String data )
	{
	user.active();
	data = data.trim();
	log( "new name: " + user.nameAtHost + " -> " + data );
	addLine( user.nameAtHost + " is now known as " + data + "." );
	user.setName( data );
	}

    private void newLine( ChatServletUser user, String data )
	{
	user.active();
	data = data.trim();
	log( "got line from " + user.nameAtHost + ": " + data );
	addLine( user.name + ": " + data );
	}

    private void disconnect( ChatServletUser user, String data )
	{
	log( "disconnected: " + user.nameAtHost );
	addLine( user.nameAtHost + " has disconnected." );
	user.connected = false;
	}

    private synchronized void addLine( String line )
	{
	lastLine = ( lastLine + 1 ) % lines.length;
	lines[lastLine] = line + "\n";
	notifyAll();
	}

    private void expire()
	{
	long now = System.currentTimeMillis();
	Enumeration en = userTable.keys();
	while ( en.hasMoreElements() )
	    {
	    String userid = (String) en.nextElement();
	    ChatServletUser user = (ChatServletUser) userTable.get( userid );
	    if ( user.connected )
		if ( now - user.lastTimeActive >= ChatUtils.INT_IDLE )
		    user.connected = false;
	    }
	}

    }


class ChatServletUser
    {

    String userid;
    String name;
    String host;
    String nameAtHost;
    int lastLineSeen;
    long lastTimeActive;
    boolean connected;

    public ChatServletUser( String userid, String name, String host, int lastLineSeen )
	{
	this.userid = userid;
	this.host = host;
	setName( name );
	this.lastLineSeen = lastLineSeen;
	active();
	}
    
    public void setName( String name )
	{
	this.name = name;
	nameAtHost = name + "@" + host;
	}
    
    public void active()
	{
	connected = true;
	lastTimeActive = System.currentTimeMillis();
	}

    }
